/*-
 * Copyright (c) 2000-2002 Andrey Simonenko
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char rcsid[] =
  "@(#)$Id: kipfil.c,v 1.4.2.8 2002/10/28 21:40:38 simon Exp $";
#endif /* !lint */

#include "system.h"

#ifdef WITH_IPFIL

#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <unistd.h>
#include <varargs.h>
#include <sys/ioctl.h>
#include <paths.h>

#include <limits.h>
#include <sys/param.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <net/if.h>

#ifdef IP_COMPAT_H
# include IP_COMPAT_H
#else
# include <netinet/ip_compat.h>
#endif

#ifdef IP_FIL_H
# include IP_FIL_H
#else
# include <netinet/ip_fil.h>
#endif

#include "kipfil.h"

#include "config.h"
#include "debug.h"


#ifdef _PATH_KMEM
# define KMEMDEV	_PATH_KMEM
#else
# define KMEMDEV	"/dev/kmem"
#endif

#ifdef IPL_NAME
# define IPLDEV IPL_NAME
#else
# define IPLDEV	"/dev/ipl"
#endif


struct kipfil	*kipfil;
u_int		nkipfil;

u_quad_t	kipfil_bcnt_max;	/* max value of IP Filter byte counter */

static const char	*kmemdev = KMEMDEV;
static const char	*ipldev = IPLDEV;

static int	kmemfd = -1, iplfd = -1;
static int	kipfil_was_changed;

static u_int	fr_que;			/* FR_INQUE or FR_OUTQUE */
static char	fil_type;		/* 'i' or 'o' */

/*
 * Wrapper for read() function.
 */
static ssize_t
Read(int fd, void *vptr, size_t n)
{
	size_t		nleft = n;
	ssize_t		nread;
	char		*ptr = vptr;

	while (nleft > 0) {
		if ( (nread = read(fd, ptr, nleft)) < 0) {
			syslog(LOG_ERR, "Read: read(FD %d): %m", fd);
			return nread;	/* error */
		} else if (nread == 0)
			break;		/* EOF */
		nleft -= nread;
		ptr += nread;
	}
	return n - nleft;
}

/*
 * Read data from /dev/kmem device.
 */
static int
kmemread(void *vptr, off_t offset, size_t n)
{
	if (lseek(kmemfd, offset, SEEK_SET) < 0) {
		syslog(LOG_ERR, "kmemread: lseek(%s, %qu)", kmemdev, offset);
		return -1;
	}
	if (Read(kmemfd, vptr, n) < 0) {
		syslog(LOG_ERR, "kmemread: Read(%s, %d)", kmemdev, n);
		return -1;
	}
	return 0;
}

/*
 * Init IP Filter support: open /dev/kmem, /dev/ipl devices.
 */
int
kipfil_init(void)
{
	char	errmsg[] = "IP Filter is not configured in the kernel or does not work properly";
	struct frentry	tmp;

	syslog(LOG_INFO, "init IP Filter support");
	if ( (kmemfd = open(kmemdev, O_RDONLY)) < 0) {
		syslog(LOG_ERR, "kipfil_init: open(%s, O_RDONLY): %m", kmemdev);
		syslog(LOG_ERR, "%s", errmsg);
		return -1;
	}
	if ( (iplfd = open(ipldev, O_RDONLY)) < 0) {
		syslog(LOG_ERR, "kipfil_init: open(%s, O_RDONLY): %m", ipldev);
		syslog(LOG_ERR, "%s", errmsg);
		return -1;
	}
	kipfil = NULL;
	nkipfil = 0;
	kipfil_was_changed = 1;

	switch (sizeof(tmp.fr_bytes)) {
	case 4:
		kipfil_bcnt_max = ULONG_MAX;
		break;
	case 8:
		kipfil_bcnt_max = UQUAD_MAX;
		break;
	default:
		syslog(LOG_ERR, "kipfil_init: unsupported size of fr_bytes (%d bytes) in struct frentry",
		    sizeof tmp.fr_bytes);
		return -1;
	}

	if (debug_ipfil > 0) {
		syslog(LOG_INFO, "kipfil_init: %s opened, FD %d", kmemdev, kmemfd);
		syslog(LOG_INFO, "kipfil_init: %s opened, FD %d", ipldev, iplfd);
		if (debug_ipfil > 1)
			syslog(LOG_INFO, "kipfil_init: sizeof(fr_bytes) = %u",
			    sizeof tmp.fr_bytes);
	}
	return 0;
}

/*
 * Free whole table with IP Filter accounting information.
 */
static void
kipfil_free_table(void)
{
	u_int		i;
	struct kipfil	*kfilp;

	if (debug_ipfil > 0)
		syslog(LOG_INFO, "kipfil_free_table: free IPFIL table");

	for (i = 0, kfilp = kipfil; i < nkipfil; ++kfilp, ++i)
		free(kfilp->bcnt);
	free(kipfil);
	kipfil = NULL;
	nkipfil = 0;
}

/*
 * Remove IP Filter support: close descriptors for opened
 * /dev/kmem and /dev/ipl devices.
 */
void
kipfil_close(void)
{
	syslog(LOG_INFO, "remove IP Filter support");
	if (kmemfd != -1) {
		if (close(kmemfd) < 0)
			syslog(LOG_ERR, "kipfil_close: close(FD %d (%s)): %m",
			    kmemfd, kmemdev);
		kmemfd = -1;
	}
	if (iplfd != -1) {
		if (close(iplfd) < 0)
			syslog(LOG_ERR, "kipfil_close close(FD %d (%s)): %m",
			    iplfd, ipldev);
		iplfd = -1;
	}
	kipfil_free_table();
}

/*
 * Dump IP Filter accounting information to syslog().
 * This funcion is called when debug_ipfil > 1.
 */
static void
kipfil_dump_table(const struct frentry *fp, u_int number, char type)
{
#define TXT_LEN 60
	static int	total_len = 0;
	static char	txt[TXT_LEN + 1];

	char	one[10 + 5 + 1 + 1 + 1 + 1];
	/* 10 - length of u_int, 10 - length of u_int, 1 - type, 1 - '@',
	   1 - ' ', 1 - '\0' */
	int	len;

	if (fp != NULL) {
		if (total_len == 0)
			syslog(LOG_INFO, "kipfil_dump_table: dump IPFIL table:");
		len = snprintf(one, sizeof(one), "%c%u@%u ", type, (u_int)fp->fr_group, number);
		if (total_len + len > TXT_LEN) {
			syslog(LOG_INFO, "  ** %s", txt);
			total_len = 0;
		}
		strcpy(txt + total_len, one);
		total_len += len;
	} else if (total_len > 0) {
		syslog(LOG_INFO, "  ** %s", txt);
		total_len = 0;
	}
}

/*
 * Store accounting information for one frentry in kipfil table.
 */
static int
frentry_store(const struct frentry *fp)
{
	int		found_flag = 0;
	u_int		i;
	size_t		size;
	u_quad_t	*bcntp;
	struct kipfil	*kfilp;

	/* Try to find entry for fp->fr_group group in our kipfil array */
	for (i = 0, kfilp = kipfil; i < nkipfil; ++kfilp, ++i) {
		if (kfilp->group == fp->fr_group && kfilp->type == fil_type) {
			found_flag = 1;
			break;
		}
		if (kipfil_was_changed == 0)
			/* If IPFIL table wasn't changed, groups are sorted.
			   Can we break loop? */
			if (kfilp->group > fp->fr_group)
				break;
	}
	if (found_flag) {
		/* Entry was found*/
		if (kfilp->next_set == kfilp->size) {
			/* New rule in group -> need to resize table */
			size = (kfilp->size + 1) * sizeof *bcntp;
			if ( (bcntp = realloc(kfilp->bcnt, size)) == NULL) {
				syslog(LOG_ERR, "frentry_store: realloc(%d bytes): %m",
				    size);
				return -1;
			}
			kfilp->bcnt = bcntp;
			++kfilp->size;
		}
	} else {
		/* Need to add new entry -> kipfil table become changed */
		size = (nkipfil + 1) * sizeof *kfilp;
		if ( (kfilp = realloc(kipfil, size)) == NULL) {
			syslog(LOG_ERR, "frentry_store: realloc(%d bytes): %m",
			    size);
			return -1;
		}
		kipfil = kfilp;
		kfilp += nkipfil;
		++nkipfil;
		kfilp->group = fp->fr_group;
		kfilp->next_set = 0;
		kfilp->type = fil_type;
		if ( (kfilp->bcnt = malloc(sizeof *kfilp->bcnt)) == NULL) {
			syslog(LOG_ERR, "frentry_store: malloc(%d bytes): %m",
			    sizeof *kfilp->bcnt);
			return -1;
		}
		kfilp->size = 1;
		kipfil_was_changed = 1;
	}
	kfilp->bcnt[kfilp->next_set] = fp->fr_bytes;

	if (debug_ipfil > 1)
		kipfil_dump_table(fp, kfilp->next_set + 1, fil_type);

	++kfilp->next_set;
	return 0;
}

/*
 * If some group was removed, then free whole table with IP Filter accounting
 * information. If number of records in some group was changed, then realloc
 * memory for this group.
 */
static int
kipfil_reset_table(void)
{
	u_int		i;
	struct kipfil	*kfilp;

	for (i = 0, kfilp = kipfil; i < nkipfil; ++kfilp, ++i) {
		if (kfilp->next_set == 0) {
			/* no records in this group */
			if (debug_ipfil > 0)
				syslog(LOG_INFO, "kipfil_reset_table: IPFIL (accounting) group %c@%u was removed from IPFIL table",
				    kfilp->type, kfilp->group);
			kipfil_free_table();
			break;
		}
		if (kfilp->next_set < kfilp->size) {
			u_quad_t	*bcntp;
			size_t		size;

			/* number of records was changed */
			if (debug_ipfil > 0)
				syslog(LOG_INFO, "kipfil_reset_table: resize IPFIL (accounting) group %c%u, old size = %u, new size = %u records",
				    kfilp->type, kfilp->type, kfilp->size, kfilp->next_set);
			size = kfilp->next_set * sizeof *bcntp;
			if ( (bcntp = realloc(kfilp->bcnt, size)) == NULL) {
				syslog(LOG_ERR, "kipfil_reset_table: realloc(%d bytes): %m", size);
				return -1;
			}
			kfilp->bcnt = bcntp;
			kfilp->size = kfilp->next_set;
		}
		kfilp->next_set = 0; /* set next free entry to first record */
	}
	return 0;
}

/*
 * Read data for group (in the first call read data for group 0) and if
 * some entry is head of group, then recursive read data for this group.
 */
static int
frentry_read(struct frentry *fr_ptr)
{
	u_int		i;
	struct frentry	fr_ent;

	for (i = 1; fr_ptr != NULL; fr_ptr = fr_ptr->fr_next, ++i) {
		if (kmemread(&fr_ent, (u_long)fr_ptr, sizeof fr_ent) < 0) {
			syslog(LOG_ERR, "frentry_read: kmemread failed for %c%u@%u, offset = %lu",
			    fr_ptr->fr_flags & FR_OUTQUE ? 'o' : fr_ptr->fr_flags & FR_INQUE ? 'i' : '?',
			    fr_ptr->fr_group, i, (u_long)fr_ptr);
			return -1;
		}
		fr_ptr = &fr_ent;
		if ((fr_ptr->fr_flags & FR_OUTQUE) == fr_que ||
		    (fr_ptr->fr_flags & FR_INQUE)  == fr_que) {
			if (frentry_store(fr_ptr) < 0)
				return -1;
		} else {
			syslog(LOG_WARNING, "frentry_read: unknown/incorrect type of frentry (fr_flags = 0x%x) for ?%u@%u, assume it as %c%u@%u",
			    (u_int)fr_ptr->fr_flags, (u_int)fr_ptr->fr_group, i,
			    fil_type, (u_int)fr_ptr->fr_group, i);
			if (frentry_store(fr_ptr) < 0)
				return -1;
		}
		if (fr_ptr->fr_grp != NULL)
			if (frentry_read(fr_ptr->fr_grp) < 0)
				return -1;
	}
	return 0;
}

/*
 * Compare two kipfil structures by group number.
 */
static int
cmp_kipfil(const void *p1, const void *p2)
{
	const struct kipfil	*q1 = p1, *q2 = p2;

	if (q1->group > q2->group)
		return 1;
	return q1->group < q2->group ? -1 : 0;
}

/*
 * Read whole IP Filter table and store it in *kipfil.
 */
int
kipfil_read_table(void)
{
	int		set;
	struct friostat	fio;
#if IPF_VERSION >= 30400
	struct friostat *fiop = &fio;
#endif
	struct frentry	*fp;
	
	if (kipfil_reset_table() < 0)
		return -1;
	bzero(&fio, sizeof fio);

#if IPF_VERSION >= 30400
	if (ioctl(iplfd, SIOCGETFS, &fiop) < 0) {
#else
	if (ioctl(iplfd, SIOCGETFS, &fio) < 0) {
#endif
		syslog(LOG_ERR, "kipfil_read_table: ioctl(FD %d (%s), SIOCGETFS)",
		    iplfd, ipldev);
		return -1;
	}

	set = fio.f_active;

	fp = fio.f_acctout[set];
	if (fp != NULL) {
		fr_que = FR_OUTQUE;
		fil_type = 'o';
		if (frentry_read(fp) < 0)
			return -1;
	}

	fp = fio.f_acctin[set];
	if (fp != NULL) {
		fr_que = FR_INQUE;
		fil_type = 'i';
		if (frentry_read(fp) < 0)
			return -1;
	}

	if (debug_ipfil > 1)
		kipfil_dump_table((struct frentry *)NULL, 0, 'X');
	if (kipfil_was_changed) {
		if (debug_ipfil > 0)
			syslog(LOG_INFO, "kipfil_read_table: IPFIL table was changed");
		qsort(kipfil, nkipfil, sizeof *kipfil, cmp_kipfil);
		kipfil_was_changed = 0;
	}
	return 0;
}

#endif /* WITH_IPFIL */
