/*-
 * Copyright (c) 1999-2002 Andrey Simonenko
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char rcsid[] =
  "@(#)$Id: rules.c,v 1.4.2.16 2002/10/28 21:40:39 simon Exp $";
#endif /* !lint */

#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/resource.h>
#include <dirent.h>
#include <unistd.h>
#include <varargs.h>
#include <regex.h>

#include "rules.h"

#include "cmd.h"
#include "common.h"
#include "config.h"
#include "db.h"
#include "debug.h"
#include "ipa.h"
#include "kipfil.h"
#ifdef __FreeBSD__
# include "kipfw.h"
# include "kip6fw.h"
#endif
#ifdef __OpenBSD__
# include "kpf.h"
#endif
#include "path.h"
#include "system.h"

struct ipac	*ipac = NULL;		/* ptr on accounting rules array */
u_int		nipac = 0;		/* #rules */

int		firstcall = 1;	/* ==0 if we pass one loop in run_ipac() */

/* use_ipfw (use_ip6fw, use_ipfil, use_pf) are equal to 1 if "ipfw"
("ip6fw", "ipfil", "pf") parameter was used in some "rule" section */
#ifdef WITH_IPFW
int		use_ipfw;
#endif
#ifdef WITH_IP6FW
int		use_ip6fw;
#endif
#ifdef WITH_IPFIL
int		use_ipfil;
#endif
#ifdef WITH_PF
int		use_pf;
#endif

int		use_worktime_global;	/* ==1 if "worktime" parameter was set in "global" section */
struct worktime	worktime_global[7];	/* global { worktime } */
struct commands	startup_global;		/* global { startup } */
struct commands	shutdown_global;	/* global { shutdown } */

static time_t	newday_time;		/* UTC time in seconds of next day */
static u_int	sleep_seconds;		/* how many seconds to sleep */
static int	process_mon = 0,	/* current month */
		process_year = 0;	/* current year */
static int	time_changed_flag = 0;	/* ==1 if there is a problem with time/localtime */

static const u_quad_t	zero_ULL = 0ULL;
static const u_quad_t	uquad_max_ULL = UQUAD_MAX;

static volatile sig_atomic_t	reconfigure_flag = 0, need_rules_check = 0,
				wakeup_flag = 0, goingdown_flag = 0,
				db_dump_flag = 0, child_terminated_flag = 0;

static int	do_reconfigure(void);

static int	free_ipac_list(struct ipac *, u_int);

static void	check_childs(void);

static time_t	check_rule_limits(const struct ipac *, time_t *);
static void	add_chunk_to_rule_limits(const struct ipac *, const u_quad_t *);
static void	sub_chunk_from_rule_limits(const struct ipac *, const u_quad_t *);

static int	Ftruncate_Seek(FILE *, long, const char *);

static int	append_limit_z_ts(struct limit *);
static int	append_limit_e_ts(struct limit *);
static int	restart_limit(const struct ipac *, struct limit *);
static int	update_rule_limits(const struct ipac *);

static time_t	mk_time_exp_upto(time_t, char);


/*
 * Add chunk to counter of *acp rule and to counters of its limits.
 * Check if there are some "subtracted" bytes. Check for counter overflowing.
 */
static int
add_chunk_to_rule(struct ipac *acp, const u_quad_t *chunk_ptr)
{
	u_quad_t	chunk = *chunk_ptr;

	if (acp->bcnt_sub >= chunk)
		acp->bcnt_sub -= chunk;
	else {
		chunk -= acp->bcnt_sub;
		acp->bcnt_sub = 0;
		if (acp->bcnt <= UQUAD_MAX - chunk)
			acp->bcnt += chunk;
		else {
			if (update_db(acp, &uquad_max_ULL) < 0)
				return -1;
			acp->bcnt = UQUAD_MAX - acp->bcnt;
			acp->bcnt = chunk - acp->bcnt;
			if (append_db(acp, &acp->bcnt) < 0)
				return -1;
		}
	}
	if (acp->limit != NULL)
		add_chunk_to_rule_limits(acp, &chunk);
	return 0;
}

#ifdef WITH_IPFW
/*
 * Do IP accounting for rule *acp from IPv4 Firewall table.
 */
static int
do_ipfwac_for_rule(struct ipac *acp)
{
	u_int		i;
	int		seen_flag;
	int		prevnum = -1;	/* real value of number is u_short type */
	u_int		subnumber = 0;	/* initial value isn't used */
	u_quad_t	chunk = 0;	/* initial value isn't used */
	u_quad_t	maxchunk = acp->maxchunk;
	struct ipfwac	*fwacp = acp->ipfwac;
	struct ip_fw	*kfwp = kipfw;

	for (i = 0; i < acp->nipfwac; ++fwacp, ++i) {
#ifdef IPFW2
		for (seen_flag = 0; !seen_flag; kfwp = (void *)kfwp + RULESIZE(kfwp)) {
#else
		for (seen_flag = 0; !seen_flag; ++kfwp) {
#endif
			if (IPFWP_NUMBER(kfwp) > fwacp->number)
				break;
			if (prevnum == IPFWP_NUMBER(kfwp))
				++subnumber;
			else {
				prevnum = IPFWP_NUMBER(kfwp);
				subnumber = 0;
			}
			if (IPFWP_NUMBER(kfwp) == fwacp->number &&
			    subnumber == fwacp->subnumber) {
				/* found IPFW rule in kernel table */
				seen_flag = 1;
				if (fwacp->set == 0) {
					fwacp->bcnt_old = IPFWP_BCNT(kfwp);
					fwacp->set = 1;
					if (!firstcall)
						syslog(LOG_WARNING, "IPFW rule %hu.%u was added to IPFW kernel table (rule %s)",
						    fwacp->number, fwacp->subnumber, acp->rulename);
				} else if (fwacp->bcnt_old != IPFWP_BCNT(kfwp)) {
					if (fwacp->bcnt_old < IPFWP_BCNT(kfwp))
						chunk = IPFWP_BCNT(kfwp) - fwacp->bcnt_old;
					else {
						chunk += kipfw_bcnt_max - (fwacp->bcnt_old - IPFWP_BCNT(kfwp));
						if (maxchunk != 0) {
							if (chunk > maxchunk) {
								syslog(LOG_WARNING, "Was IPFW rule %hu.%u flushed? Rule %s maxchunk = %s, current chunk = %s",
								    fwacp->number, fwacp->subnumber, acp->rulename,
								    show_size(maxchunk), show_size2(chunk));
								chunk = IPFWP_BCNT(kfwp);
								syslog(LOG_INFO, "Add absolute value %s of IPFW rule %hu.%u counter to rule %s",
								    show_size(chunk), fwacp->number, fwacp->subnumber, acp->rulename);
							}
						} else
							syslog(LOG_WARNING, "Counter of IPFW rule %hu.%u was overflowed. Previous counter = %s, current counter = %s",
							    fwacp->number, fwacp->subnumber, show_size(fwacp->bcnt_old), show_size2(IPFWP_BCNT(kfwp)));
					}
					fwacp->bcnt_old = IPFWP_BCNT(kfwp);
					if (fwacp->action == ADD) {
						if (add_chunk_to_rule(acp, &chunk) < 0)
							return -1;
					} else /* fwacp->action == SUB */ {
						if (acp->bcnt >= chunk)
							acp->bcnt -= chunk;
						else {
							if (acp->bcnt_sub <= UQUAD_MAX - (chunk - acp->bcnt))
								acp->bcnt_sub += chunk - acp->bcnt;
							else {
								syslog(LOG_ERR, "rule %s: wrong configuration for IPFW rule %hu.%u, counter overflowed",
								    acp->rulename, fwacp->number, fwacp->subnumber);
								acp->bcnt_sub = UQUAD_MAX;
							}
							acp->bcnt = 0;
						}
						if (acp->limit != NULL)
							sub_chunk_from_rule_limits(acp, &chunk);
					}
				}
			}
			if (IPFWP_NUMBER(kfwp) == IPFW_NUMBER_MAX)
				break;
		}
		if (!seen_flag) {
			if (fwacp->set != 0) {
				fwacp->set = 0;
				syslog(LOG_WARNING, "IPFW rule %hu.%u was removed from IPFW kernel table (rule %s)",
				    fwacp->number, fwacp->subnumber, acp->rulename);
			} else if (firstcall)
				syslog(LOG_WARNING, "IPFW rule %hu.%u does not exist in IPFW kernel table (rule %s)",
				    fwacp->number, fwacp->subnumber, acp->rulename);
		}
	}
	if (update_db(acp, &acp->bcnt) < 0)
		return -1;
	return update_rule_limits(acp);
}
#endif /* WITH_IPFW */

#ifdef WITH_IP6FW
/*
 * Do IP accounting for rule *acp from IPv6 Firewall table.
 */
static int
do_ip6fwac_for_rule(struct ipac *acp)
{
	u_int		i;
	int		seen_flag;
	int		prevnum = -1;	/* real value of number is u_short type */
	u_int		subnumber = 0;	/* initial value isn't used */
	u_quad_t	chunk = 0;	/* initial value isn't used */
	u_quad_t	maxchunk = acp->maxchunk;
	struct ipfwac	*fwacp = acp->ip6fwac;
	struct ip6_fw	*kfwp = kip6fw;

	for (i = 0; i < acp->nip6fwac; ++fwacp, ++i) {
		for (seen_flag = 0; !seen_flag; ++kfwp) {
			if (kfwp->fw_number > fwacp->number)
				break;
			if (prevnum == kfwp->fw_number)
				++subnumber;
			else {
				prevnum = kfwp->fw_number;
				subnumber = 0;
			}
			if (kfwp->fw_number == fwacp->number &&
			    subnumber == fwacp->subnumber) {
				/* found IP6FW rule in kernel table */
				seen_flag = 1;
				if (fwacp->set == 0) {
					fwacp->bcnt_old = kfwp->fw_bcnt;
					fwacp->set = 1;
					if (!firstcall)
						syslog(LOG_WARNING, "IP6FW rule %hu.%u was added to IP6FW kernel table (rule %s)",
						    fwacp->number, fwacp->subnumber, acp->rulename);
				} else if (fwacp->bcnt_old != kfwp->fw_bcnt) {
					if (fwacp->bcnt_old < kfwp->fw_bcnt)
						chunk = kfwp->fw_bcnt - fwacp->bcnt_old;
					else {
						chunk += kip6fw_bcnt_max - (fwacp->bcnt_old - kfwp->fw_bcnt);
						if (maxchunk != 0) {
							if (chunk > maxchunk) {
								syslog(LOG_WARNING, "Was IP6FW rule %hu.%u flushed? Rule %s maxchunk = %s, current chunk = %s",
								    fwacp->number, fwacp->subnumber, acp->rulename,
								    show_size(maxchunk), show_size2(chunk));
								chunk = kfwp->fw_bcnt;
								syslog(LOG_INFO, "Add absolute value %s of IP6FW rule %hu.%u counter to rule %s",
								    show_size(chunk), fwacp->number, fwacp->subnumber, acp->rulename);
							}
						} else
							syslog(LOG_WARNING, "Counter of ip6fw rule %hu.%u was overflowed. Previous counter = %s, current counter = %s",
							    fwacp->number, fwacp->subnumber, show_size(fwacp->bcnt_old), show_size2(kfwp->fw_bcnt));
					}
					fwacp->bcnt_old = kfwp->fw_bcnt;
					if (fwacp->action == ADD) {
						if (add_chunk_to_rule(acp, &chunk) < 0)
							return -1;
					} else /* fwacp->action == SUB */ {
						if (acp->bcnt >= chunk)
							acp->bcnt -= chunk;
						else {
							if (acp->bcnt_sub <= UQUAD_MAX - (chunk - acp->bcnt))
								acp->bcnt_sub += chunk - acp->bcnt;
							else {
								syslog(LOG_ERR, "rule %s: wrong configuration for IP6FW rule -%hu.%u, counter overflowed",
								    acp->rulename, fwacp->number, fwacp->subnumber);
								acp->bcnt_sub = UQUAD_MAX;
							}
							acp->bcnt = 0;
						}
						if (acp->limit != NULL)
							sub_chunk_from_rule_limits(acp, &chunk);
					}
				}
			}
			if (kfwp->fw_number == IP6FW_NUMBER_MAX)
				break;
		}
		if (!seen_flag) {
			if (fwacp->set != 0) {
				fwacp->set = 0;
				syslog(LOG_WARNING, "ip6fw rule %hu.%u was removed from IP6FW kernel table (rule %s)",
				    fwacp->number, fwacp->subnumber, acp->rulename);
			} else if (firstcall)
				syslog(LOG_WARNING, "IP6FW rule %hu.%u does not exist in IP6FW kernel table (rule %s)",
				    fwacp->number, fwacp->subnumber, acp->rulename);
		}
	}
	if (update_db(acp, &acp->bcnt) < 0)
		return -1;
	return update_rule_limits(acp);
}
#endif /* WITH_IP6FW */

#ifdef WITH_IPFIL
/*
 * Do IP accounting for rule *acp from IP Filter accounting table.
 */
static int
do_ipfilac_for_rule(struct ipac *acp)
{
	u_int		i, j;
	u_int		k = 0;		/* initial value isn't used */
	int		seen_group_flag;
	u_quad_t	kbcnt;
	u_quad_t	chunk = 0;	/* initial value isn't used */
	u_quad_t	maxchunk = acp->maxchunk;
	struct ipfilac		*filacp;
	struct ipfilac_list	*list = NULL; /* initial value isn't used */
	struct kipfil		*kfilp = kipfil;

	for (i = j = 0, filacp = acp->ipfilac; i < acp->nipfilac; ++filacp, ++i) {
		for (seen_group_flag = 0; j < nkipfil; ++kfilp, ++j) {
			if (kfilp->group > filacp->group || seen_group_flag)
				break;
			if (kfilp->group == filacp->group &&
			    kfilp->type == filacp->type) {
				/* found IPFIL group in IPFIL kernel table */
				seen_group_flag = 1;
				for (k = 0, list = filacp->list; k < filacp->nlist; ++list, ++k) {
					if (list->number < kfilp->size + 1) {
						kbcnt = kfilp->bcnt[list->number - 1];
						if (list->set == 0) {
							list->bcnt_old = kbcnt;
							list->set = 1;
							if (!firstcall)
								syslog(LOG_WARNING, "IPFIL accounting rule %c%u@%u was added to the IPFIL kernel table (rule %s)",
								    filacp->type, filacp->group, list->number, acp->rulename);
						} else {
							if (list->bcnt_old != kbcnt) {
								if (list->bcnt_old < kbcnt)
									chunk = kbcnt - list->bcnt_old;
								else {
									chunk += kipfil_bcnt_max - (list->bcnt_old - kbcnt);
									if (maxchunk != 0) {
										if (chunk > maxchunk) {
											syslog(LOG_WARNING, "Was IPFIL rule %c%u@%u flushed? Rule %s maxchunk = %s, current chunk = %s",
											    filacp->type, filacp->group, list->number, acp->rulename,
											    show_size(maxchunk), show_size2(chunk));
											chunk = kbcnt;
											syslog(LOG_INFO, "Add absolute value %s of IPFIL rule %c%u@%u counter to rule %s",
											    show_size(chunk), filacp->type, filacp->group, list->number, acp->rulename);
										}
									} else
										syslog(LOG_WARNING, "Counter of IPFIL rule %c%u@%u was overflowed. Previous counter = %s, current counter = %s",
										    filacp->type, filacp->group, list->number, show_size(list->bcnt_old), show_size2(kbcnt));
								}
								list->bcnt_old = kbcnt;
								if (list->action == ADD) {
									if (add_chunk_to_rule(acp, &chunk) < 0)
										return -1;
								} else /* list->action == SUB */ {
									if (acp->bcnt >= chunk)
										acp->bcnt -= chunk;
									else {
										if (acp->bcnt_sub <= UQUAD_MAX - (chunk - acp->bcnt))
											acp->bcnt_sub += chunk - acp->bcnt;
										else {
											syslog(LOG_ERR, "rule %s: wrong configuration for the IPFIL rule %c%u@%u, counter overflowed",
											    acp->rulename, filacp->type, filacp->group, list->number);
											acp->bcnt_sub = UQUAD_MAX;
										}
										acp->bcnt = 0;
									}
									if (acp->limit != NULL)
										sub_chunk_from_rule_limits(acp, &chunk);
								}
							}
						}
					} else
						/* some of ipfil rules haven't been found */
						break;
				}
			}
		}
		if (!seen_group_flag) {
			k = 0;
			list = filacp->list;
		}
		if ((seen_group_flag && list->number >= kfilp->size + 1) || !seen_group_flag) {
			for (; k < filacp->nlist; ++list, ++k) {
				if (list->set != 0) {
					list->set = 0;
					syslog(LOG_WARNING, "IPFIL accounting rule %c%u@%u was removed from IPFIL kernel table (rule %s)",
					    filacp->type, filacp->group, list->number, acp->rulename);
				} else if (firstcall)
					syslog(LOG_WARNING, "IPFIL accounting rule %c%u@%u does not exist in IPFIL kernel table (rule %s)",
					    filacp->type, filacp->group, list->number, acp->rulename);
			}
		}
	}
	if (update_db(acp, &acp->bcnt) < 0)
		return -1;
	return update_rule_limits(acp);
}
#endif /* WITH_IPFIL */

#ifdef WITH_PF
/*
 * Do IP accounting for rule *acp from Packet Filter table.
 */
static int
do_pfac_for_rule(struct ipac *acp)
{
	u_int		i, nr;
	u_quad_t	chunk = 0;	/* initial value isn't used */
	u_quad_t	maxchunk = acp->maxchunk;
	struct pfac	*pfacp = acp->pfac;
	u_quad_t	*kpfp;

	for (i = 0; i < acp->npfac; ++pfacp, ++i) {
		nr = pfacp->number;
		if (nr < nkpf) {
			/* found PF rule in the kernel table */
			kpfp = kpf + nr;
			if (pfacp->set == 0) {
				pfacp->bcnt_old = *kpfp;
				pfacp->set = 1;
				if (!firstcall)
					syslog(LOG_WARNING, "PF rule %u was added to PF kernel table (rule %s)",
					    nr, acp->rulename);
			} else if (pfacp->bcnt_old != *kpfp) {
				if (pfacp->bcnt_old < *kpfp)
					chunk = *kpfp- pfacp->bcnt_old;
				else {
					chunk += UQUAD_MAX - (pfacp->bcnt_old - *kpfp);
					if (maxchunk != 0) {
						if (chunk > maxchunk) {
							syslog(LOG_WARNING, "Was PF rule %u flushed? Rule %s maxchunk = %s, current chunk = %s",
							    nr, acp->rulename, show_size(maxchunk), show_size2(chunk));
							chunk = *kpfp;
							syslog(LOG_INFO, "Add absolute value %s of PF rule %u counter to rule %s",
							    show_size(chunk), nr, acp->rulename);
						}
					} else
						syslog(LOG_WARNING, "Counter of PF rule %u was overflowed. Previous counter = %s, current counter = %s",
						    nr, show_size(pfacp->bcnt_old), show_size2(*kpfp));
				}
				pfacp->bcnt_old = *kpfp;
				if (pfacp->action == ADD) {
					if (add_chunk_to_rule(acp, &chunk) < 0)
						return -1;
				} else /* pfacp->action == SUB */ {
					if (acp->bcnt >= chunk)
						acp->bcnt -= chunk;
					else {
						if (acp->bcnt_sub <= UQUAD_MAX - (chunk - acp->bcnt))
							acp->bcnt_sub += chunk - acp->bcnt;
						else {
							syslog(LOG_ERR, "rule %s: wrong configuration for PF rule %u, counter overflowed",
							    acp->rulename, nr);
							acp->bcnt_sub = UQUAD_MAX;
						}
						acp->bcnt = 0;
					}
					if (acp->limit != NULL)
						sub_chunk_from_rule_limits(acp, &chunk);
				}
			}
		} else {
			if (pfacp->set != 0) {
				pfacp->set = 0;
				syslog(LOG_WARNING, "PF rule %u was removed from PF kernel table (rule %s)",
				    nr, acp->rulename);
			} else if (firstcall)
				syslog(LOG_WARNING, "PF rule %u does not exist in PF kernel table (rule %s)",
				    nr, acp->rulename);
		}
	}
	if (update_db(acp, &acp->bcnt) < 0)
		return -1;
	return update_rule_limits(acp);
}
#endif /* WITH_PF */

/*
 * Make accounting for all rules, check limits, calculate new sleep time.
 */
static int
do_ipac(void)
{
#define UPDATE_SLEEP_SECONDS()				\
	rule_sleep_seconds = acp->wakeup - curr_time;	\
	if (sleep_seconds > rule_sleep_seconds)		\
		sleep_seconds = rule_sleep_seconds

	u_int		i, rule_sleep_seconds;
	time_t		wakeup_limits;
	struct ipac	*acp;

	sleep_seconds = DAY;	/* 1 day in seconds */
	for (i = 0, acp = ipac; i < nipac; ++acp, ++i) {
		if (acp->wakeup <= curr_time || need_rules_check) {
#ifdef WITH_IPFW
			if (acp->ipfwac != NULL)
				if (do_ipfwac_for_rule(acp) < 0)
					return -1;
#endif

#ifdef WITH_IP6FW
			if (acp->ip6fwac != NULL)
				if (do_ip6fwac_for_rule(acp) < 0)
					return -1;
#endif

#ifdef WITH_IPFIL
			if (acp->ipfilac != NULL)
				if (do_ipfilac_for_rule(acp) < 0)
					return -1;
#endif

#ifdef WITH_PF
			if (acp->pfac != NULL)
				if (do_pfac_for_rule(acp) < 0)
					return -1;
#endif

			if (acp->append_db_time != 0)
				if (acp->newrec_time <= curr_time && curr_tm.tm_hour != 24) {
					/* append new record, but do not append it at the end of day */
					if (append_db(acp, &zero_ULL) < 0)
						return -1;
					acp->bcnt = 0;
				}
			acp->wakeup = curr_time + acp->update_db_time;
			rule_sleep_seconds = acp->update_db_time;
			if (sleep_seconds > acp->update_db_time)
				sleep_seconds = rule_sleep_seconds;
			if (acp->append_db_time != 0)
				if (acp->wakeup > acp->newrec_time) {
					acp->wakeup = acp->newrec_time;
					UPDATE_SLEEP_SECONDS();
				}
			if (acp->limit != NULL) {
				if (check_rule_limits(acp, &wakeup_limits) < 0)
					return -1;
				if (wakeup_limits > 0 && acp->wakeup > wakeup_limits) {
					acp->wakeup = wakeup_limits;
					UPDATE_SLEEP_SECONDS();
				}
			}
			if (debug_time > 1)
				syslog(LOG_INFO, "do_ipac: rule %s, new sleep = %u sec.", acp->rulename, rule_sleep_seconds);
		} else {
			UPDATE_SLEEP_SECONDS();
			if (debug_time > 1)
				syslog(LOG_INFO, "do_ipac: rule %s, rest sleep = %u sec.", acp->rulename, rule_sleep_seconds);
		}
	}
	need_rules_check = 0;
	return 0;
}

/*
 * New day came, close old database files if needed, else open new ones.
 */
static int
newday_came(void)
{
	int		curr_year = curr_tm.tm_year + 1900,
			curr_mon = curr_tm.tm_mon + 1;
	u_int		i;
	char		*path;
	char		oneline[DBRECORDSIZE + 1];
	struct stat	statbuf;
	struct ipac	*acp;
	struct tm	db_tm1, db_tm2;
	u_quad_t	db_bcnt; /* really we are not interesting in its value */

	for (i = 0, acp = ipac; i < nipac; ++acp, ++i) {
		acp->bcnt = 0ULL; /* flush counter */

		if (process_mon == curr_mon && process_year == curr_year) {
			/* new day came, but the same month,
			   just add entry with 0 bytes for new day */
			if (append_db(acp, &zero_ULL) < 0)
				return -1;
			continue;
		}
		/* new month came */
		if (acp->fp != NULL) {
			/* close file for previous month */
			if (fclose(acp->fp) != 0) {
				syslog(LOG_ERR, "fclose(%s): %m", acp->filename);
				return -1;
			}
			free(acp->filename);
		}
		if (asprintf(&path, "%s/%s/%d", db_dir, acp->rulename, curr_year) < 0) {
			syslog(LOG_ERR, "asprintf: %m");
			return -1;
		}
		switch (check_db_dir(path, acp)) {
		case 1:	/* dir is present */
			break;
		case 0: /* no such dir */
			/* create directory DBDIR/<rule>/YYYY */
			if (create_db_dir(path, acp) < 0)
				return -1;
			break;
		default: /* -1, error */
			return -1;
		}
		free(path);

		if (asprintf(&path, "%s/%s/%d/%02d", db_dir, acp->rulename, curr_year, curr_mon) < 0) {
			syslog(LOG_ERR, "asprintf: %m");
			return -1;
		}
		switch (check_db_file(path, acp)) {
		case 1: /* file is present */
			break;
		case 0: /* no such file */
			/* new month came, create new file and append record
			   with 0 bytes */
			if ( (acp->fp = create_db_file(path, acp)) == NULL)
				return -1;
			acp->fd = fileno(acp->fp);
			acp->filename = path;
			if (append_db(acp, &zero_ULL) < 0)
				return -1;
			continue;
		default: /* -1, error */
			return -1;
		}
		/* new month came, but file for it already exists */
		if (firstcall == 0)
			/* do not say following warning just after start */
			syslog(LOG_WARNING, "new month %d/%02d came, but file %s for it already exists",
			    curr_year, curr_mon, path);
		if ( (acp->fp = fopen(path, "r+")) == NULL) {
			syslog(LOG_ERR, "fopen(%s, \"r+\"): %m", path);
			return -1;
		}
		if (lstat(path, &statbuf) < 0) {
			syslog(LOG_ERR, "lstat(%s): %m", path);
			return -1;
		}
		acp->fd = fileno(acp->fp);
		acp->filename = path;
		if (statbuf.st_size < DBRECORDSIZE) {
			syslog(LOG_ERR, "wrong format of file %s: too small file size", acp->filename);
			return -1;
		}
		if (fseek(acp->fp, (long)-DBRECORDSIZE, SEEK_END) < 0) {
			syslog(LOG_ERR, "fseek(%s, -%d, SEEK_END): %m", acp->filename, DBRECORDSIZE);
			return -1;
		}
		if (fread(oneline, sizeof(char), DBRECORDSIZE, acp->fp) != DBRECORDSIZE) {
			syslog(LOG_ERR, "fread(%s): %m", acp->filename);
			return -1;
		}
		oneline[DBRECORDSIZE] = '\0';
		if (REGEXEC(db_entry, oneline) != 0) {
			syslog(LOG_ERR, "wrong format of file %s: cannot recognize last record", acp->filename);
			return -1;
		}
		memcpy(&db_tm1, &curr_tm, sizeof db_tm1); /* set year, month, TZ, etc. */
		memcpy(&db_tm2, &curr_tm, sizeof db_tm2); /* set year, month, TZ, etc. */
		errno = 0;
		if (sscanf(oneline, "%d/%d:%d:%d-%d:%d:%d %qu", &db_tm1.tm_mday,
		    &db_tm1.tm_hour, &db_tm1.tm_min, &db_tm1.tm_sec,
		    &db_tm2.tm_hour, &db_tm2.tm_min, &db_tm2.tm_sec,
		    &db_bcnt) != 8) {
			syslog(LOG_ERR, "%s: sscanf(%s, %%d/%%d:%%d:%%d-%%d:%%d:%%d %%qu): failed: %m",
			    acp->filename, oneline);
			return -1;
		}
		db_tm2.tm_mday = db_tm1.tm_mday;
		++db_tm1.tm_mon;
		++db_tm2.tm_mon;
		if (check_date(&db_tm1) < 0 ||
		    check_date(&db_tm2) < 0 ||
		    tmcmp(&db_tm1, &db_tm2) > 0) {
			syslog(LOG_ERR, "wrong format of file %s: wrong timestamp in last record", acp->filename);
			return -1;
		}
		--db_tm1.tm_mon;
		--db_tm2.tm_mon;
		if (tmcmp(&db_tm2, &curr_tm) > 0)
			syslog(LOG_WARNING, "file %s: last record is more fresh than new one", acp->filename);
		if (append_db(acp, &zero_ULL) < 0)
			return -1;
	}
	process_mon = curr_tm.tm_mon + 1;
	process_year = curr_tm.tm_year + 1900;
	/* get time UTC of new day in seconds */
	newday_time = mk_time_exp_upto(curr_time, UPTO_DAY);
	return 0;
}

/*
 * The main function for accounting, is called from main().
 */
int
run_ipac(void)
{
	time_t		new_curr_time;
	struct tm	new_curr_tm;
	static sigset_t	sigmask, zeromask, pendmask;

	sigemptyset(&zeromask);
	sigemptyset(&sigmask);

	if (debug)
		sigaddset(&sigmask, SIGINT);
	sigaddset(&sigmask, SIGTERM);
	sigaddset(&sigmask, SIGHUP);
	sigaddset(&sigmask, SIGCHLD);
	sigaddset(&sigmask, SIGUSR1);

	if (sigprocmask(SIG_BLOCK, &sigmask, (sigset_t *)NULL) < 0) {
		syslog(LOG_ERR, "sigprocmask(SIG_BLOCK): %m");
		return -1;
	}
	/*
	 * XXX: signals SIGTERM (and SIGINT), SIGHUP, SIGUSR1 before function
	 *      sigprocmask() are not handled safe.
	 */
	syslog(LOG_INFO, "make IP accounting...");

	if (newday_came() < 0)
		return -1;

	for (;;) {
#ifdef WITH_IPFW
		if (use_ipfw)
			if (kipfw_read_table() < 0)
				return -1;
#endif

#ifdef WITH_IP6FW
		if (use_ip6fw)
			if (kip6fw_read_table() < 0)
				return -1;
#endif

#ifdef WITH_IPFIL
		if (use_ipfil)
			if (kipfil_read_table() < 0)
				return -1;
#endif

#ifdef WITH_PF
		if (use_pf)
			if (kpf_read_table() < 0)
				return -1;
#endif

		time(&new_curr_time);
		localtime_r(&new_curr_time, &new_curr_tm);

		/* 
		 * Check for problems with time and local date,
		 * XXX: if TZ is changed outside, this fact isn't reflected
		 *	on running process [is it bug of system or standard?].
		 */
		if (new_curr_time < curr_time) {
/*
 * XXX use %ld for FreeBSD, NetBSD and %d for OpenBSD to just
 *     suppress warning messages during compilation.
 */
#if defined(__FreeBSD__) || defined(__NetBSD__)
			syslog(LOG_WARNING, "time UTC goes back? old time = %lds UTC, new time = %lds UTC (delta = -%s)",
			    curr_time, new_curr_time, show_time(curr_time - new_curr_time));
#else
			syslog(LOG_WARNING, "time UTC goes back? old time = %ds UTC, new time = %ds UTC (delta = -%s)",
			    curr_time, new_curr_time, show_time(curr_time - new_curr_time));
#endif
			time_changed_flag = 1;
		}
		if ((curr_tm.tm_zone && !new_curr_tm.tm_zone) ||
		    (!curr_tm.tm_zone && new_curr_tm.tm_zone) ||
		    (curr_tm.tm_zone && new_curr_tm.tm_zone && strcmp(curr_tm.tm_zone, new_curr_tm.tm_zone) != 0)) {
			syslog(LOG_WARNING, "time zone was changed, old TZ = %s, new TZ = %s",
			    curr_tm.tm_zone ? curr_tm.tm_zone : "???",
			    new_curr_tm.tm_zone ? new_curr_tm.tm_zone : "???");
			time_changed_flag = 1;
		}
		if (curr_tm.tm_gmtoff != new_curr_tm.tm_gmtoff) {
			syslog(LOG_WARNING, "offset from UTC was changed, old gmtoff = %lds, new gmtoff = %lds (delta = %s%s)",
			    curr_tm.tm_gmtoff, new_curr_tm.tm_gmtoff,
			    curr_tm.tm_gmtoff > new_curr_tm.tm_gmtoff ? "" : "-",
			    curr_tm.tm_gmtoff > new_curr_tm.tm_gmtoff ? show_time(curr_tm.tm_gmtoff - new_curr_tm.tm_gmtoff) : show_time(new_curr_tm.tm_gmtoff - curr_tm.tm_gmtoff));
			time_changed_flag = 1;
		}
		if (curr_tm.tm_isdst != new_curr_tm.tm_isdst) {
			syslog(LOG_WARNING, "\"summer time in effect\" was changed, old isdst = %s, new isdst = %s",
			    curr_tm.tm_isdst ? "yes" : "no", new_curr_tm.tm_isdst ? "yes" : "no");
			time_changed_flag = 1;
		}
		if (tmcmp(&new_curr_tm, &curr_tm) < 0) {
			syslog(LOG_WARNING, "local time goes back? old localtime = %d.%02d.%02d/%02d:%02d:%02d (%s), new localtime = %d.%02d.%02d/%02d:%02d:%02d (%s)",
			    curr_tm.tm_year + 1900, curr_tm.tm_mon + 1, curr_tm.tm_mday, curr_tm.tm_hour, curr_tm.tm_min, curr_tm.tm_sec,
			    curr_tm.tm_zone ? curr_tm.tm_zone : "???",
			    new_curr_tm.tm_year + 1900, new_curr_tm.tm_mon + 1, new_curr_tm.tm_mday, new_curr_tm.tm_hour, new_curr_tm.tm_min, new_curr_tm.tm_sec,
			    new_curr_tm.tm_zone ? new_curr_tm.tm_zone : "???");
			time_changed_flag = 1;
		}

		curr_time = new_curr_time;

		if (time_changed_flag) {
			syslog(LOG_WARNING, "unsuccessive date or time changes are detected");
			syslog(LOG_WARNING, "flush all counters and limits with old timestamp");
			/* 
			 * XXX store new accounting with old timestamp,
			 *     curr_tm isn't changed and is equal to old curr_tm.
			 */
			need_rules_check = 1;
			if (do_ipac() < 0)
				return -1;
			syslog(LOG_WARNING, "append new records for all rules");
			memcpy(&curr_tm, &new_curr_tm, sizeof curr_tm);
			process_mon = process_year = 0;
			if (newday_came() < 0)
				return -1;
			if (init_limits(0) < 0)
				return -1;
			time_changed_flag = 0;
			continue;
		}

		if (curr_time >= newday_time) {
			if (new_curr_tm.tm_mday != curr_tm.tm_mday) {
				need_rules_check = 1;
				/* update record at the end of day with
				   second timestamp equal to 24:00:00 */
				curr_tm.tm_hour = 24;
				curr_tm.tm_min = curr_tm.tm_sec = 0;
				if (do_ipac() < 0)
					return -1;
				memcpy(&curr_tm, &new_curr_tm, sizeof curr_tm);
				/* some time can be spent in do_ipac(), so we should
                                   set first timestamp of new day as 00:00:00, this isn't
				   correct but will not confused user runs ipastat(8) with
				   the -t switch */
				curr_tm.tm_hour = curr_tm.tm_min = curr_tm.tm_sec = 0;
				if (newday_came() < 0)
					return -1;
			} else
				syslog(LOG_WARNING, "new day did not come, but was expected");
			memcpy(&curr_tm, &new_curr_tm, sizeof curr_tm);
			process_mon = curr_tm.tm_mon + 1;
			process_year = curr_tm.tm_year + 1900;
			/* get time UTC of new day in seconds */
			new_curr_tm.tm_hour = new_curr_tm.tm_min = new_curr_tm.tm_sec = 0;
			newday_time = mktime(&new_curr_tm) + DAY; 
		} else 
			memcpy(&curr_tm, &new_curr_tm, sizeof curr_tm);

		if (do_ipac() < 0)
			return -1;
		firstcall = 0;
		if (goingdown_flag)
			return 0;
		if (reconfigure_flag) {
			switch (do_reconfigure()) {
			case 0:
				/* configuration file was successfuly parsed */
				syslog(LOG_INFO, "make IP accounting...");
				if (newday_came() < 0)
					return -1;
				reconfigure_flag = 0;
				continue;
			case -1:
				/* configuration file was successful parsed,
				   but k*_init() failed */
				return -1;
			/* do_reconfigure() also returns -2, which means that
			   there are problems with configuration file(s)
			   parsing */
			}
			reconfigure_flag = 0;
		}
		if (db_dump_flag) {
			syslog(LOG_INFO, "accounting information was successfuly dumped to database");
			db_dump_flag = 0;
		}

		if (sleep_seconds + curr_time > newday_time)
			sleep_seconds = newday_time - curr_time;

		if (debug_time > 0)
			syslog(LOG_INFO, "run_ipac: sleep_seconds = %u sec.", sleep_seconds);

		if (sleep_seconds != 0) {
			alarm(sleep_seconds);	/* set alarm */
			while (wakeup_flag == 0)
				sigsuspend(&zeromask);
			alarm(0);		/* release alarm */
		} else {
			/* should not set alarm, also signals INT, TERM, HUP,
			   CHLD and USR1 should be tested */
			if (sigpending(&pendmask) < 0) {
				syslog(LOG_ERR, "sigpending: %m");
				return -1;
			}
			if (sigismember(&pendmask, SIGTERM) ||
			    sigismember(&pendmask, SIGHUP) ||
			    sigismember(&pendmask, SIGCHLD) ||
			    sigismember(&pendmask, SIGUSR1) ||
			    (debug && sigismember(&pendmask, SIGINT)))
				sigsuspend(&zeromask);	/* should be interrupted immediately */
		}
		if (child_terminated_flag) {
			check_childs();
			child_terminated_flag = 0;
		}
		if (reconfigure_flag)
			syslog(LOG_WARNING, "caught signal %d '%s', reconfiguring...",
			    SIGHUP, sys_signame[SIGHUP]);
		if (db_dump_flag)
			syslog(LOG_INFO, "caught signal %d '%s', dumping database...",
			    SIGUSR1, sys_signame[SIGUSR1]);
		if (goingdown_flag)
			syslog(LOG_INFO, "caught signal %d '%s', shutdowning...",
			    goingdown_flag, sys_signame[goingdown_flag]);
		wakeup_flag = 0;
	}
}

/*
 * Allocate memory for new struct ipac in *acp list, set initial
 * values for its fields.
 */
struct ipac *
ipac_new(void)
{
	u_int		i;
	struct ipac	*acp;

	if ( (acp = realloc(ipac, (nipac + 1) * sizeof *acp)) == NULL) {
		syslog(LOG_ERR, "ipac_new: realloc: %m");
		return NULL;
	}
	ipac = acp;
	acp += nipac;
	++nipac;
	acp->fp = NULL;
	acp->bcnt_sub = 0;
	acp->rulename = NULL;
	acp->filename = NULL;
#ifdef WITH_IPFW
	acp->ipfwac = NULL;
#endif
#ifdef WITH_IP6FW
	acp->ip6fwac = NULL;
#endif
#ifdef WITH_IPFIL
	acp->ipfilac = NULL;
#endif
#ifdef WITH_PF
	acp->pfac = NULL;
#endif
	acp->limit = NULL;
	acp->rc[0].cmd = acp->rc[0].cmd_if_limit = acp->rc[0].cmd_if_nolimit = NULL;
	acp->rc[1].cmd = acp->rc[1].cmd_if_limit = acp->rc[1].cmd_if_nolimit = NULL;
	acp->info = NULL;
	acp->maxchunk = 0;
	acp->wakeup = 0;
	for (i = 0; i < 7; ++i) {
		acp->worktime[i].interval = NULL;
		acp->worktime[i].current = 0;
	}
	return acp;
}

/*
 * Append new time stamp in limit file.
 */
static int
append_limit_ts(const struct limit *lp, char type, const struct tm *tm_ptr)
{
	if (fseek(lp->fp, 0L, SEEK_END) < 0) {
		syslog(LOG_ERR, "fseek(%s, 0, SEEK_END): %m", lp->limitname);
		return -1;
	}
	if (fprintf(lp->fp, "%c %d.%02d.%02d/%02d:%02d:%02d\n", type,
	    tm_ptr->tm_year + 1900, tm_ptr->tm_mon + 1, tm_ptr->tm_mday,
	    tm_ptr->tm_hour, tm_ptr->tm_min, tm_ptr->tm_sec) < LIMIT_TIMESTAMP_SIZE) {
		syslog(LOG_ERR, "fprintf(%s): %m", lp->limitname);
		return -1;
	}
	if (fflush(lp->fp) != 0) {
		syslog(LOG_ERR, "fflush(%s): %m", lp->filename);
		return -1;
	}
	return 0;
}

/*
 * SIGCHLD signal handler.
 */
void
sig_chld(int signo)
{
	child_terminated_flag = wakeup_flag = 1;
}

/*
 * SIGINT and/or SIGTERM signal(s) handler.
 */
void
end_work(int signo)
{
	need_rules_check = wakeup_flag = 1;
	goingdown_flag = signo;
}

/*
 * SIGALRM signal handler, can be caught from alarm() in run_ipac().
 */
void
sig_alrm(int signo)
{
	wakeup_flag = 1;
}

/*
 * SIGUSR1 signal handler.
 */
void
force_db_dump(int signo)
{
	db_dump_flag = need_rules_check = wakeup_flag = 1;
}

/*
 * SIGHUP signal handler, cause reconfiguring.
 */
void
reconfigure(int signo)
{
	reconfigure_flag = need_rules_check = wakeup_flag = 1;
}

/*
 * Reread configuration file, parse it and if it is correct, then use
 * new settings.
 */
static int
do_reconfigure(void)
{
	int		i;
	struct ipac	*ipac_bkp = ipac;
	u_int		nipac_bkp = nipac;
#ifdef WITH_IPFW
	int		use_ipfw_bkp = use_ipfw;
#endif
#ifdef WITH_IP6FW
	int		use_ip6fw_bkp = use_ip6fw;
#endif
#ifdef WITH_IPFIL
	int		use_ipfil_bkp = use_ipfil;
#endif
#ifdef WITH_PF
	int		use_pf_bkp = use_pf;
#endif
	int		lock_db_bkp = lock_db;
	u_int		lock_wait_time_bkp = lock_wait_time;
	char		*db_dir_bkp = db_dir;
	int		use_worktime_global_bkp = use_worktime_global;
	struct worktime	worktime_global_bkp[7];
	struct commands	shutdown_global_bkp = shutdown_global;

	syslog(LOG_INFO, "rereading configuration file...");
	ipac = NULL;
	nipac = 0;
	backup_debug();
	for (i = 0; i < 7; ++i)
		memcpy(&worktime_global_bkp[i], &worktime_global[i], sizeof(struct worktime));

	if (parse_config(RECONFIG_PARSING) < 0) {
		syslog(LOG_WARNING, "continue using previous configuration");
		if (free_ipac_list(ipac, nipac) < 0)
			return -1;
		ipac = ipac_bkp;
		nipac = nipac_bkp;
#ifdef WITH_IPFW
		use_ipfw = use_ipfw_bkp;
#endif
#ifdef WITH_IP6FW
		use_ip6fw = use_ip6fw_bkp;
#endif
#ifdef WITH_IPFIL
		use_ipfil = use_ipfil_bkp;
#endif
#ifdef WITH_PF
		use_pf = use_pf_bkp;
#endif
		db_dir = db_dir_bkp;
		lock_db = lock_db_bkp;
		lock_wait_time = lock_wait_time_bkp;
		shutdown_global = shutdown_global_bkp;
		for (i = 0; i < 7; ++i)
			memcpy(&worktime_global[i], &worktime_global_bkp[i], sizeof(struct worktime));
		use_worktime_global = use_worktime_global_bkp;
		restore_debug();
		return -2;
	}

#ifdef WITH_IPFW
	if (use_ipfw_bkp && !use_ipfw)
		kipfw_close();
	else if (!use_ipfw_bkp && use_ipfw)
		if (kipfw_init() < 0)
			return -1;
#endif

#ifdef WITH_IP6FW
	if (use_ip6fw_bkp && !use_ip6fw)
		kip6fw_close();
	else if (!use_ip6fw_bkp && use_ip6fw)
		if (kip6fw_init() < 0)
			return -1;
#endif

#ifdef WITH_IPFIL
	if (use_ipfil_bkp && !use_ipfil)
		kipfil_close();
	else if (!use_ipfil_bkp && use_ipfil)
		if (kipfil_init() < 0)
			return -1;
#endif

#ifdef WITH_PF
	if (use_pf_bkp && !use_pf)
		kpf_close();
	else if (!use_pf_bkp && use_pf)
		if (kpf_init() < 0)
			return -1;
#endif

	syslog(LOG_INFO, "use new configuration");
	if (free_ipac_list(ipac_bkp, nipac_bkp) < 0)
		return -1;
	free_cmd_list(shutdown_global.cmd, shutdown_global.ncmd);
	for (i = 0; i < 7; ++i)
		free(worktime_global_bkp[i].interval);
	process_mon = process_year = 0;
	firstcall = 1;
	return init_db();
}

/*
 * Free one ipac structure.
 */
int
free_ipac(struct ipac *acp)
{
#ifdef WITH_IPFIL
	struct ipfilac	*ipfilacp;
#endif
	struct limit	*lp;
	u_int		j, k;

	free(acp->rulename);
	if (acp->fp != NULL)
		if (fclose(acp->fp) != 0) {
			syslog(LOG_ERR, "fclose(%s): %m", acp->filename);
			return -1;
		}
	free(acp->filename);
#ifdef WITH_IPFW
	free(acp->ipfwac);
#endif
#ifdef WITH_IP6FW
	free(acp->ip6fwac);
#endif
#ifdef WITH_IPFIL
	if (acp->ipfilac != NULL) {
		for (ipfilacp = acp->ipfilac, j = 0; j < acp->nipfilac; ++ipfilacp, ++j)
			free(ipfilacp->list);
		free(acp->ipfilac);
	}
#endif
#ifdef WITH_PF
	free(acp->pfac);
#endif
	for (k = 0; k < 7; ++k)
		free(acp->worktime[k].interval);
	if (acp->limit != NULL) {
		for (lp = acp->limit, j = 0; j < acp->nlimit; ++lp, ++j) {
			free(lp->limitname);
			if (lp->fp != NULL) {
				if (fclose(lp->fp) != 0) {
					syslog(LOG_ERR, "fclose(%s): %m", lp->filename);
					return -1;
				}
				free(lp->filename);
			}
			free_cmd_list(lp->reach.cmd, lp->reach.ncmd);
			free_cmd_list(lp->expire.cmd, lp->expire.ncmd);
			free_commands(&lp->rc[0]);
			free_commands(&lp->rc[1]);
			for (k = 0; k < 7; ++k)
				free(lp->worktime[k].interval);
		}
		free(acp->limit);
	}
	free_commands(&acp->rc[0]);
	free_commands(&acp->rc[1]);
	return 0;
}

/*
 * Free whole *ipac_head list.
 */
static int
free_ipac_list(struct ipac *ipac_head, u_int n)
{
	struct ipac	*acp;
	u_int		i;

	for (acp = ipac_head, i = 0; i < n; ++acp, ++i)
		if (free_ipac(acp) < 0)
			return -1;
	free(ipac_head);
	return 0;
}

/*
 * Update byte counters in limits for rule *acp.
 */
static int
update_rule_limits(const struct ipac *acp)
{
	u_int		i;
	struct limit	*lp;

	for (i = 0, lp = acp->limit; i < acp->nlimit; ++lp, ++i)
		if (lp->is_changed) {
			if (fseek(lp->fp, 0L, SEEK_SET) < 0) {
				syslog(LOG_ERR, "fseek(%s, 0, SEEK_SET): %m", lp->filename);
				return -1;
			}
			if (lock_db_file_until_end(lp->filename, lp->fd, SEEK_SET) < 0)
				return -1;
			if (fprintf(lp->fp, "%020qu", lp->bcnt) != LIMIT_COUNTER_SIZE) {
				syslog(LOG_ERR, "fprintf(%s, \"%%020qu\"), failed: %m", lp->filename);
				return -1;
			}
			if (fflush(lp->fp) != 0) {
				syslog(LOG_ERR, "fflush(%s): %m", lp->filename);
				return -1;
			}
			if (unlock_db_file(lp->filename, lp->fd) < 0)
				return -1;
			lp->is_changed = 0;
		}
	return 0;
}

/*
 * Add chunk bytes to every not reached limit for rule *acp.
 */
static void
add_chunk_to_rule_limits(const struct ipac *acp, const u_quad_t *chunk_ptr)
{
	u_int		i;
	u_quad_t	chunk = *chunk_ptr;
	struct limit	*lp;

	if (chunk == 0)
		return;
	for (i = 0, lp = acp->limit; i < acp->nlimit; ++lp, ++i) {
		if (lp->status & LIMIT_IS_REACHED)
			continue;
		if (lp->bcnt_sub >= chunk)
			lp->bcnt_sub -= chunk;
		else {
			chunk -= lp->bcnt_sub;
			lp->bcnt_sub = 0;
			if (lp->bcnt <= UQUAD_MAX - chunk)
				lp->bcnt += chunk;
			else {
				lp->bcnt = UQUAD_MAX;
				syslog(LOG_WARNING, "rule %s, limit %s: byte counter overflowed",
				    acp->rulename, lp->limitname);
			}
			lp->is_changed = 1;
		}
	}
}

/*
 * Subtract chunk byte from every not reached limit for rule *acp.
 */
static void
sub_chunk_from_rule_limits(const struct ipac *acp, const u_quad_t *chunk_ptr)
{
	u_int		i;
	struct limit	*lp;

	for (i = 0, lp = acp->limit; i < acp->nlimit; ++lp, ++i) {
		if (lp->status & LIMIT_IS_REACHED)
			continue;
		if (lp->bcnt >= *chunk_ptr)
			lp->bcnt -= *chunk_ptr;
		else {
			if (lp->bcnt_sub <= UQUAD_MAX - (*chunk_ptr - lp->bcnt))
				lp->bcnt_sub += *chunk_ptr - lp->bcnt;
			else {
				syslog(LOG_WARNING, "rule %s, limit %s: wrong configuration for some accounting rule",
				    acp->rulename, lp->limitname);
				lp->bcnt_sub = UQUAD_MAX;
			}
			lp->bcnt = 0;
		}
		lp->is_changed = 1;
	}
}

/*
 * Run commands in "startup" and "shutdown" sections in "rule" sections.
 * If x == 0 then run commands from "startup" sections, else from
 * "shutdown" ones.
 */
int
run_rules_rc(int x)
{
	u_int		i, j;
	int		limit_is_reached;
	char		*msg = x == 0 ? "startup" : "shutdown";
	struct ipac	*acp;
	struct limit	*lp;
	
	for (i = 0, acp = ipac; i < nipac; ++acp, ++i) {
		if (acp->rc[x].cmd != NULL) {
			syslog(LOG_INFO, "rule %s (%s): run commands", acp->rulename, msg);
			if (exec_cmd_list(acp->rc[x].cmd, acp->rc[x].ncmd,
			    "rule %s (%s)", acp->rulename, msg) < 0)
				return -1;
		}
		limit_is_reached = 0;
		if (acp->rc[x].cmd_if_limit != NULL || acp->rc[x].cmd_if_nolimit != NULL) {
			limit_is_reached = 0;
			for (j = 0, lp = acp->limit; j < acp->nlimit; ++lp, ++j)
				if (lp->bcnt >= lp->byte_limit) {
					limit_is_reached = 1;
					break;
				}
			if (limit_is_reached && acp->rc[x].cmd_if_limit != NULL) {
				syslog(LOG_INFO, "rule %s (%s, if_limit): run commands", acp->rulename, msg);
				if (exec_cmd_list(acp->rc[x].cmd_if_limit, acp->rc[x].ncmd_if_limit,
				    "rule %s (%s, if_limit)", acp->rulename, msg) < 0)
				return -1;
			} else if (!limit_is_reached && acp->rc[x].cmd_if_nolimit != NULL) {
				syslog(LOG_INFO, "rule %s (%s, if_nolimit): run commands",
				    acp->rulename, msg);
				if (exec_cmd_list(acp->rc[x].cmd_if_nolimit, acp->rc[x].ncmd_if_nolimit,
				    "rule %s (%s, if_nolimit)", acp->rulename, msg) < 0)
					return -1;
			}
		}
		for (j = 0, lp = acp->limit; j < acp->nlimit; ++lp, ++j) {
			if (lp->rc[x].cmd != NULL) {
				syslog(LOG_INFO, "rule %s, limit %s (%s): run commands",
				    acp->rulename, lp->limitname, msg);
				if (exec_cmd_list(lp->rc[x].cmd, lp->rc[x].ncmd,
				    "rule %s, limit %s (%s)", acp->rulename, lp->limitname, msg) < 0)
					return -1;
			}
			if (lp->status & LIMIT_IS_REACHED) {
				if (lp->rc[x].cmd_if_limit != NULL) {
					syslog(LOG_INFO, "rule %s, limit %s (%s, if_limit): run commands",
					    acp->rulename, lp->limitname, msg);
					if (exec_cmd_list(lp->rc[x].cmd_if_limit, lp->rc[x].ncmd_if_limit,
					    "rule %s, limit %s (%s, if_limit)", acp->rulename, lp->limitname, msg) < 0)
						return -1;
				}
			} else if (lp->rc[x].cmd_if_nolimit != NULL) {
				syslog(LOG_INFO, "rule %s, limit %s (%s, if_nolimit): run commands",
				    acp->rulename, lp->limitname, msg);
				if (exec_cmd_list(lp->rc[x].cmd_if_nolimit, lp->rc[x].ncmd_if_nolimit,
				    "rule %s, limit %s (%s, if_nolimit)", acp->rulename, lp->limitname, msg) < 0)
					return -1;
			}
		}
	}
	return 0;
}

/*
 * This routine implemets "limit" section.
 */
static time_t
check_rule_limits(const struct ipac *acp, time_t *wakeup)
{
	u_int		i;
	struct limit	*lp;

	*wakeup = curr_time + DAY; /* just initial value */
	for (i = 0, lp = acp->limit; i < acp->nlimit; ++lp, ++i) {
		if (lp->status & LIMIT_IS_REACHED) {
			/* limit has been already reached */
			if (lp->status & LIMIT_EXEC)
				continue; /* some commands are run in background */
			if (lp->expire.time.upto != UPTO_NOTSET) {
				/* "expire" section is used */
				if (!(lp->status & LIMIT_EXPIRE_CHECKED)) {
					/* first time we come here */
					if (lock_db_file_until_end(lp->filename, lp->fd, SEEK_SET) < 0)
						return -1;
					if (append_limit_e_ts(lp) < 0)
						return -1;
					if (unlock_db_file(lp->filename, lp->fd) < 0)
						return -1;
					lp->status |= LIMIT_EXPIRE_CHECKED;
					if (*wakeup > lp->expire_time)
						*wakeup = lp->expire_time;
				} else if (lp->expire_time <= curr_time) {
					/* it's time to expire limit */
					if (lp->expire.cmd != NULL && !(lp->status & LIMIT_EXPIRE_EXEC)) {
						/* "exec" parameter in the "expire" section is used and
						    we haven't run any command from the "expire" section yet */
						if (debug_limit > 0)
							syslog(LOG_INFO, "rule %s, limit %s: limit expired",
							    acp->rulename, lp->limitname);
						if (debug_exec > 0)
							syslog(LOG_INFO, "rule %s, limit %s (expire): run commands",
							    acp->rulename, lp->limitname);
						if ( (lp->pid = exec_cmd_list_bg(lp->expire.cmd, lp->expire.ncmd, "rule %s, limit %s (expire)",
						    acp->rulename, lp->limitname)) < 0)
							return -1;
						if (lock_db_file_until_end(lp->filename, lp->fd, SEEK_SET) < 0)
							return -1;
						if (append_limit_ts(lp, LIMIT_EXECUTED, &curr_tm) < 0)
							return -1;
						if (unlock_db_file(lp->filename, lp->fd) < 0)
							return -1;
						lp->status |= LIMIT_EXPIRE_EXEC | LIMIT_EXEC;
					} else if (lp->expire.cmd == NULL ||
						   (lp->expire.cmd != NULL && (lp->status & LIMIT_EXPIRE_EXEC))) {
						/* "exec" parameters in "expire" section are not used,
						   or are used, but commands already have been executed */
						if (lp->expire.cmd == NULL && debug_limit > 0)
							syslog(LOG_INFO, "rule %s, limit %s: limit expired",
							    acp->rulename, lp->limitname);
						if (lock_db_file_until_end(lp->filename, lp->fd, SEEK_SET) < 0)
							return -1;
						if (restart_limit(acp, lp) < 0)
							return -1;
						if (unlock_db_file(lp->filename, lp->fd) < 0)
							return -1;
						if (lp->zero_time_param.upto != UPTO_NOTSET)
							if (*wakeup > lp->zero_time)
								*wakeup = lp->zero_time;
						lp->status = 0;
					}
				} else {
					/* just check time to expire */
					if (*wakeup > lp->expire_time)
						*wakeup = lp->expire_time;
				}
			}
		} else {
			/* limit hasn't been reached before */
			if (lp->bcnt >= lp->byte_limit) {
				/* limit has just been reached */
				if (debug_limit > 0)
					syslog(LOG_INFO, "rule %s, limit %s: limit reached, bcnt = %s",
					    acp->rulename, lp->limitname, show_size(lp->bcnt));
				lp->status = LIMIT_IS_REACHED;
				lp->reach_time = curr_time;
				if (lock_db_file_until_end(lp->filename, lp->fd, SEEK_SET) < 0)
					return -1;
				if (append_limit_ts(lp, LIMIT_REACHED, &curr_tm) < 0)
					return -1;
				if (lp->reach.cmd != NULL) {
					/* "exec" parameter in "reach" section is used */
					if (debug_exec > 0)
						syslog(LOG_INFO, "rule %s, limit %s (reach): run commands", 
						    acp->rulename, lp->limitname);
					if ( (lp->pid = exec_cmd_list_bg(lp->reach.cmd, lp->reach.ncmd, "rule %s, limit %s (reach)",
					    acp->rulename, lp->limitname)) < 0)
						return -1;
					if (append_limit_ts(lp, LIMIT_EXECUTED, &curr_tm) < 0)
						return -1;
					lp->status |= LIMIT_EXEC;
				} else if (lp->expire.time.upto != UPTO_NOTSET) {
					if (append_limit_e_ts(lp) < 0)
						return -1;
					lp->status |= LIMIT_EXPIRE_CHECKED;
					if (*wakeup > lp->expire_time)
						*wakeup = lp->expire_time;
				}
				if (unlock_db_file(lp->filename, lp->fd) < 0)
					return -1;
			} else if (lp->zero_time_param.upto != UPTO_NOTSET) {
				/* "zero_time" parameter in "limit" section is used */
				if (lp->zero_time <= curr_time) {
					/* it is time to restart limit */
					if (debug_limit > 0)
						syslog(LOG_INFO, "rule %s, limit %s: limit zeroed",
						    acp->rulename, lp->limitname);
					if (lock_db_file_until_end(lp->filename, lp->fd, SEEK_SET) < 0)
						return -1;
					if (restart_limit(acp, lp) < 0)
						return -1;
					if (unlock_db_file(lp->filename, lp->fd) < 0)
						return -1;
				}
				if (*wakeup > lp->zero_time)
					*wakeup = lp->zero_time;
			}
		}
	}
	return 0;
}

static time_t
mk_time_exp_upto(time_t start_time, char upto)
{
	time_t		result_time = start_time;
	struct tm	tmp_tm;

	if (upto == UPTO_SIMPLE)
		return start_time;

	localtime_r(&result_time, &tmp_tm);

	if (upto == UPTO_MONTH) {
		/* up to the end of month */
		if (tmp_tm.tm_mon == 11) {
			tmp_tm.tm_mon = 0;
			++tmp_tm.tm_year;
		} else
			++tmp_tm.tm_mon;
		tmp_tm.tm_mday = 1;
		tmp_tm.tm_hour = tmp_tm.tm_min = tmp_tm.tm_sec = 0;
		return mktime(&tmp_tm);
	}
	/* up to the end of minute */
	result_time += MINUTE - tmp_tm.tm_sec;
	if (upto == UPTO_MINUTE)
		return result_time;
	++tmp_tm.tm_min;
	
	/* up to the end of hour */
	result_time += MINUTE * (60 - tmp_tm.tm_min);
	if (upto == UPTO_HOUR)
		return result_time;
	++tmp_tm.tm_hour;
	
	/* up to the end of day */
	result_time += HOUR * (24 - tmp_tm.tm_hour);
	if (upto == UPTO_DAY)
		return result_time;
	++tmp_tm.tm_wday;

	/* up to the end of week */
	if (tmp_tm.tm_wday != 1)
		result_time += DAY * (8 - tmp_tm.tm_wday);
	return result_time;	/* upto == UPTO_WEEK */
}

/*
 * Return calculated time for "+X" like time.
 */
static time_t
mk_time_exp(time_t start_time, const struct time_exp *texp)
{
	time_t		result_time;
	u_int		monthes, years;
	struct tm	result_tm;

	result_time = start_time;
	if (texp->side == 0)
		result_time = mk_time_exp_upto(start_time, texp->upto);
	result_time += texp->seconds;
	if (texp->monthes != 0) {
		localtime_r(&result_time, &result_tm);
		if (texp->monthes != 0) {
			if (result_tm.tm_mon + texp->monthes > 11) {
				monthes = texp->monthes - (11 - result_tm.tm_mon);
				years = monthes / 12;
				result_tm.tm_year += 1 + years;
				result_tm.tm_mon = monthes - 12 * years - 1;
			} else
				result_tm.tm_mon += texp->monthes;
			if (result_tm.tm_mday > 28) {
				switch (result_tm.tm_mon) {
				case 3:
				case 5:
				case 8:
				case 10:
					if (result_tm.tm_mday > 30)
						result_tm.tm_mday = 30;
					break;
				case 1:
					result_tm.tm_mday = result_tm.tm_year % 4 == 0 ? 29 : 28;
				}
			}
		}
		result_time = mktime(&result_tm);
	}
	return texp->side == 0 ? result_time : mk_time_exp_upto(result_time, texp->upto);
}

/*
 * Restart limit *lp, append 's' and 'z' timestamp if needed.
 */
static int
restart_limit(const struct ipac *acp, struct limit *lp)
{
	if (debug_limit > 0)
		syslog(LOG_INFO, "rule %s, limit %s: restart limit",
		    acp->rulename, lp->limitname);
	if (Ftruncate_Seek(lp->fp, 0L, lp->filename) < 0)
		return -1;
	if (fprintf(lp->fp, "%020qu %020qu\n", 0ULL, lp->byte_limit) != LIMIT_LINE1_SIZE) {
		syslog(LOG_ERR, "fprintf(%s, \"%%020qu %%020qu\\n\"), failed: %m", lp->filename);
		return -1;
	}
	if (append_limit_ts(lp, LIMIT_STARTED, &curr_tm) < 0)
		return -1;
	lp->start_time = curr_time;
	lp->bcnt = 0;
	return append_limit_z_ts(lp);
}

/*
 * Just send message to syslog, that error in limit was fixed.
 */
static void
log_limit_fix(struct ipac *acp, struct limit *lp)
{
	syslog(LOG_WARNING, "rule %s, limit %s: errors fixed, restart limit",
	    acp->rulename, lp->limitname);
}

/*
 * Use curr_tm and curr_time to determine if timestamps for limit are correct.
 * Limits files were opened in init_db_rules() function.
 */
int
init_limits(int mode)
{
	char		*buf = NULL;
	size_t		bufsize;
	u_int		i, j, k;
	int		len;
	u_quad_t	old_size;
	struct ipac	*acp;
	struct limit	*lp;
	struct timestamp {
		short		ts_is_set;
		long		ts_offset;
		struct tm	ts_tm;
	}		ts_start, ts_zero, ts_reach, ts_expire, ts_exec1,
			ts_exec2, *ts_ptr;

	for (i = 0, acp = ipac; i < nipac; ++acp, ++i) {
		for (j = 0, lp = acp->limit; j < acp->nlimit; ++j, ++lp) {
			if (fseek(lp->fp, 0L, SEEK_SET) < 0) {
				syslog(LOG_ERR, "init_limits: fseek(%s): %m", lp->filename);
				return -1;
			}

			if (mode == 0)
				if (lock_db_file_until_end(lp->filename, lp->fd, SEEK_SET) < 0)
					return -1;

			len = readline(&buf, &bufsize, lp->fp, lp->filename);
			if (len == 0 && feof(lp->fp) != 0) {
				/* file has been just created or it is empty */
				if (restart_limit(acp, lp) < 0)
					return -1;
				continue;
			}
			if (len == 0) {
				/* XXX readline() can't return 0 without EOF */
				syslog(LOG_ERR, "read 0 bytes from file %s (first line)", lp->filename);
				return -1;
			}
			if (len < 0) {
				syslog(LOG_ERR, "cannot read first line from file %s", lp->filename);
				return -1;
			}
			if (REGEXEC(limit_line1, buf) != 0) {
				syslog(LOG_ERR, "%s: cannot recognize first line", lp->filename);
				return -1;
			}
			errno = 0;
			if (sscanf(buf, "%qu %qu", &lp->bcnt, &old_size) != 2) {
				syslog(LOG_ERR, "%s: sscanf(%s, \"%%qu %%qu\"): failed: %m", lp->filename, buf);
				return -1;
			}
			if (lp->byte_limit != old_size)
				syslog(LOG_WARNING, "rule %s, limit %s: parameter \"byte_limit\" was changed, old size = %s new size = %s",
				    acp->rulename, lp->limitname, show_size(old_size), show_size2(lp->byte_limit));

			ts_start.ts_is_set = ts_zero.ts_is_set = ts_reach.ts_is_set = 
			    ts_expire.ts_is_set = ts_exec1.ts_is_set = ts_exec2.ts_is_set = 0;
			for (k = 0; k < 7; ++k) {
				len = readline(&buf, &bufsize, lp->fp, lp->filename);
				if (len == 0 && feof(lp->fp) != 0) {
					if (k == 0) {
						syslog(LOG_ERR, "%s: too small file size", lp->filename);
						log_limit_fix(acp, lp);
						if (restart_limit(acp, lp) < 0)
							return -1;
						goto next_limit;
					}
					break;
				}
				if (len > 0 && k == 6) {
					syslog(LOG_ERR, "%s: too big file size (more than 6 timestamps)", lp->filename);
					return -1;
				}
				if (len == 0) {
					/* XXX readline() can't return 0 without EOF */
					syslog(LOG_ERR, "%s: read 0 bytes (%d line)", lp->filename, k + 2);
					return -1;
				}
				if (len < 0) {
					syslog(LOG_ERR, "%s: cannot read %d line from", lp->filename, k + 2);
					return -1;
				}
				if (REGEXEC(limit_timestamp, buf) != 0) {
					syslog(LOG_ERR, "%s: cannot recognize %d line", lp->filename, k + 2);
					return -1;
				}
				ts_ptr = NULL;
				switch (*buf) {
				case LIMIT_STARTED:
					if (k != 0)
						break;
					if (!ts_start.ts_is_set)
						ts_ptr = &ts_start;
					break;
				case LIMIT_ZEROED:
					if (k != 1)
						break;
					if (!ts_zero.ts_is_set)
						ts_ptr = &ts_zero;
					break;
				case LIMIT_REACHED:
					if ((ts_zero.ts_is_set && k != 2) ||
					    (!ts_zero.ts_is_set && k != 1))
						break;
					if (!ts_reach.ts_is_set)
						ts_ptr = &ts_reach;
					break;
				case LIMIT_EXECUTED:
					if (k == 2 || k == 3) {
						if (!ts_exec1.ts_is_set)
							ts_ptr = &ts_exec1;
					} else if (k == 4 || k == 5)
						if (!ts_exec2.ts_is_set)
							ts_ptr = &ts_exec2;
					break;
				case LIMIT_EXPIRED:
					if ((!ts_zero.ts_is_set && !ts_exec1.ts_is_set && k != 2) ||
					    (!ts_zero.ts_is_set && ts_exec1.ts_is_set && k != 3) ||
					    (ts_zero.ts_is_set && !ts_exec1.ts_is_set && k != 3) ||
					    (ts_zero.ts_is_set && ts_exec1.ts_is_set && k != 4))
						break;
					if (!ts_expire.ts_is_set)
						ts_ptr = &ts_expire;
					break;
				default:
					syslog(LOG_ERR, "init_limits: recognized but unknown type of limit timestamp: `%c'", *buf);
					return -1;
				}
				if (ts_ptr == NULL) {
					syslog(LOG_ERR, "%s: wrong order of timestamps", lp->filename);
					return -1;
				}
				ts_ptr->ts_is_set = 1;
				if ( (ts_ptr->ts_offset = ftell(lp->fp)) < 0) {
					syslog(LOG_ERR, "ftell(%s): %m", lp->filename);
					return -1;
				}
				memcpy(&ts_ptr->ts_tm, &curr_tm, sizeof curr_tm); /* init TZ parameters, etc. */
				errno = 0;
				if (sscanf(buf + 2, "%d.%d.%d/%d:%d:%d", &ts_ptr->ts_tm.tm_year,
				    &ts_ptr->ts_tm.tm_mon, &ts_ptr->ts_tm.tm_mday, &ts_ptr->ts_tm.tm_hour,
				    &ts_ptr->ts_tm.tm_min, &ts_ptr->ts_tm.tm_sec) != 6) {
					syslog(LOG_ERR, "%s: sscanf(\"%s\", \"%%d.%%d.%%d/%%d:%%d:%%d\"): failed: %m", lp->filename, buf + 2);
					return -1;
				}
				if (check_date(&ts_ptr->ts_tm) < 0) {
					syslog(LOG_ERR, "%s: wrong values of date in %d line", lp->filename, k + 2);
					return -1;
				}
				ts_ptr->ts_tm.tm_year -= 1900;
				--ts_ptr->ts_tm.tm_mon;
			}

			if (!ts_start.ts_is_set) {
				syslog(LOG_ERR, "%s: first timestamp must be marked by `s'", lp->filename);
				return -1;
			}

			if (tmcmp(&ts_start.ts_tm, &curr_tm) > 0) {
				syslog(LOG_WARNING, "rule %s, limit %s: start_time = %d.%02d.%02d/%02d:%02d:%02d is greater than current time",
				    acp->rulename, lp->limitname,
				    ts_start.ts_tm.tm_year + 1900, ts_start.ts_tm.tm_mon + 1, ts_start.ts_tm.tm_mday,
				    ts_start.ts_tm.tm_hour, ts_start.ts_tm.tm_min, ts_start.ts_tm.tm_sec);
				log_limit_fix(acp, lp);
				if (restart_limit(acp, lp) < 0)
					return -1;
			}

			lp->start_time = mktime(&ts_start.ts_tm);

			if (lp->bcnt < old_size) {
				/* limit isn't reached with old "size" */
				if (ts_reach.ts_is_set || ts_expire.ts_is_set ||
				    ts_exec1.ts_is_set || ts_exec2.ts_is_set) {
					syslog(LOG_ERR, "%s: limit is not reached with old size, wrong timestamps were found",
					    lp->filename);
					return -1;
				}
				if (mode == 1 && ts_zero.ts_is_set) {
					if (tmcmp(&curr_tm, &ts_zero.ts_tm) > 0) {
						if (debug_limit > 0) {
							syslog(LOG_INFO, "init_limits: rule %s, limit %s: old zero_time = %d.%02d.%02d/%02d:%02d:%02d is less than current time",
							    acp->rulename, lp->limitname,
							    ts_zero.ts_tm.tm_year + 1900, ts_zero.ts_tm.tm_mon + 1, ts_zero.ts_tm.tm_mday,
							    ts_zero.ts_tm.tm_hour, ts_zero.ts_tm.tm_min, ts_zero.ts_tm.tm_sec);
						}
						if (restart_limit(acp, lp) < 0)
							return -1;
						continue;
					}
				}
				if (lp->byte_limit != old_size) {
					if (fseek(lp->fp, 0L, SEEK_SET) < 0) {
						syslog(LOG_ERR, "fseek(%s, 0, SEEK_SET): %m", lp->filename);
						return -1;
					}
					if (fprintf(lp->fp, "%020qu %020qu\n", lp->bcnt, lp->byte_limit) != LIMIT_LINE1_SIZE) {
						syslog(LOG_ERR, "init_limits: fprintf(%s, \"%%020qu %%020qu\\n\"), failed: %m", lp->filename);
						return -1;
					}
				}
				if (Ftruncate_Seek(lp->fp, ts_start.ts_offset, lp->filename) < 0)
					return -1;
				if (append_limit_z_ts(lp) < 0)
					return -1;
				if (mode == 1 && lp->zero_time_param.upto != UPTO_NOTSET) {
					if (curr_time > lp->zero_time) {
						struct tm	tmp_tm;

						localtime_r(&lp->zero_time, &tmp_tm);
						if (debug_limit > 0) {
							syslog(LOG_INFO, "init_limits: rule %s, limit %s: recalculated zero_time = %d.%02d.%02d/%02d:%02d:%02d is less than current time",
							    acp->rulename, lp->limitname,
							    tmp_tm.tm_year + 1900, tmp_tm.tm_mon + 1, tmp_tm.tm_mday,
							    tmp_tm.tm_hour, tmp_tm.tm_min, tmp_tm.tm_sec);
						}
						if (restart_limit(acp, lp) < 0)
							return -1;
						continue;
					}
				}
			} else {
				/* limit is reached with old "size" */
				if (!ts_reach.ts_is_set) {
					syslog(LOG_ERR, "file %s: limit is reached with old size, cannot find reach_time line",
					    lp->filename);
					continue;
				}
				lp->reach_time = mktime(&ts_reach.ts_tm);
				if (mode == 1 && ts_expire.ts_is_set) {
					if (tmcmp(&curr_tm, &ts_expire.ts_tm) > 0) {
						if (debug_limit > 0) {
							syslog(LOG_INFO, "init_limits: rule %s, limit %s: old expire_time = %d.%02d.%02d/%02d:%02d:%02d less than current time",
							    acp->rulename, lp->limitname,
							    ts_expire.ts_tm.tm_year + 1900, ts_expire.ts_tm.tm_mon + 1, ts_expire.ts_tm.tm_mday,
							    ts_expire.ts_tm.tm_hour, ts_expire.ts_tm.tm_min, ts_expire.ts_tm.tm_sec);
						}
						if (restart_limit(acp, lp) < 0)
							return -1;
						continue;
					}
				}
				if (ts_exec1.ts_is_set) {
					if (Ftruncate_Seek(lp->fp, ts_exec1.ts_offset, lp->filename) < 0)
						return -1;
				} else {
					if (Ftruncate_Seek(lp->fp, ts_reach.ts_offset, lp->filename) < 0)
						return -1;
				}
				if (append_limit_e_ts(lp) < 0)
					return -1;
				if (mode == 1 && lp->expire.time.upto != UPTO_NOTSET) {
					if (curr_time > lp->expire_time) {
						struct tm	tmp_tm;

						localtime_r(&lp->expire_time, &tmp_tm);
						if (debug_limit > 0) {
							syslog(LOG_INFO, "init_limits: rule %s, limit %s: recalculated expire_time = %d.%02d.%02d/%02d:%02d:%02d is less than current time",
							    acp->rulename, lp->limitname,
							    tmp_tm.tm_year + 1900, tmp_tm.tm_mon + 1, tmp_tm.tm_mday,
							    tmp_tm.tm_hour, tmp_tm.tm_min, tmp_tm.tm_sec);
						}
						if (restart_limit(acp, lp) < 0)
							return -1;
						continue;
					}
				}
				lp->status = LIMIT_IS_REACHED | LIMIT_EXPIRE_CHECKED;
			}
next_limit:
			if (mode == 0)
				if (unlock_db_file(lp->filename, lp->fd) < 0)
					return -1;
		}
	}
	free(buf);
	return 0;
}

/*
 * ftruncate() file and seek() to the end of file *fp.
 */
static int
Ftruncate_Seek(FILE *fp, long length, const char *filename)
{
	if (fflush(fp) != 0) {
		syslog(LOG_ERR, "fflush(%s): %m", filename);
		return -1;
	}
	if (ftruncate(fileno(fp), length) < 0) {
		syslog(LOG_ERR, "ftruncate(%s, %ld): %m", filename, length);
		return -1;
	}
	if (fflush(fp) != 0) {
		syslog(LOG_ERR, "fflush(%s): %m", filename);
		return -1;
	}
	if (fseek(fp, length, SEEK_SET) < 0) {
		syslog(LOG_ERR, "fseek(%s, %ld, SEEK_SET): %m", filename, length);
		return -1;
	}
	return 0;
}

/*
 * Some of child processes exited.
 */
static void
check_childs(void)
{
	u_int		i, j;
	int		status, foundchildpid = 0;
	pid_t		childpid;
	struct ipac	*acp;
	struct limit	*lp;

	while ( (childpid = waitpid(0, &status, WNOHANG)) > 0) {
		foundchildpid = 0;
		for (i = 0, acp = ipac; i < nipac; ++acp, ++i) {
			for (j = 0, lp = acp->limit; j < acp->nlimit; ++lp, ++j)
				if (lp->pid == childpid) {
					lp->pid = 0;
					lp->status ^= LIMIT_EXEC;
					acp->wakeup = 0; /* <-- this rule should be checked */
					if (debug_exec >= 3)
						syslog(LOG_INFO, "check_childs: rule %s, limit %s: PID %d terminated",
						    acp->rulename, lp->limitname, childpid);
					foundchildpid = 1;
					break;
				}
			if (foundchildpid)
				break;
		}
		if (!foundchildpid)
			syslog(LOG_WARNING, "check_childs: waitpid returned PID %d, such PID is not expected",
			    childpid);
	}
	if (!foundchildpid && childpid < 0 && errno != EINTR)
		syslog(LOG_WARNING, "check_childs: waitpid: %m");
}

/*
 * Append 'z' timestamp to limit *lp.
 */
static int
append_limit_z_ts(struct limit *lp)
{
	if (lp->zero_time_param.upto != UPTO_NOTSET) {
		struct tm	tmp_tm;

		lp->zero_time = mk_time_exp(lp->start_time, &lp->zero_time_param);
		localtime_r(&lp->zero_time, &tmp_tm);
		if (append_limit_ts(lp, LIMIT_ZEROED, &tmp_tm) < 0)
			return -1;
	}
	return 0;
}

/*
 * Append 'e' timestamp to limit *lp.
 */
static int
append_limit_e_ts(struct limit *lp)
{
	if (lp->expire.time.upto != UPTO_NOTSET) {
		struct tm	tmp_tm;

		lp->expire_time = mk_time_exp(lp->reach_time, &lp->expire.time);
		localtime_r(&lp->expire_time, &tmp_tm);
		if (append_limit_ts(lp, LIMIT_EXPIRED, &tmp_tm) < 0)
			return -1;
	}
	return 0;
}
